library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.vme64x_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity bridge_serdes_spartan6 is
  generic (
    g_mode     : t_bridge_serdes_mode := MASTER;
    g_SERDES_RX_DELAY_TAPS : integer:= 0;
    g_CLOCK_PERIOD : integer := 16
    );

  port (
    clk_125m_i : in std_logic := '0';
    rst_n_i    : in std_logic := '0';

    bridge_d_i     : in  std_logic_vector(15 downto 0);
    bridge_frame_i : in  std_logic;
    bridge_d_o     : out std_logic_vector(15 downto 0);
    bridge_frame_o : out std_logic;

    bridge_clk_o : out std_logic;
    bridge_clk2_o : out std_logic;
    bridge_rst_n_o : out std_logic;
    
    ser_clk_p_i : in std_logic := '0';
    ser_clk_n_i : in std_logic := '0';

    ser_clk_p_o : out std_logic;
    ser_clk_n_o : out std_logic;

    ser_tx_o : out std_logic_vector(4 downto 0);
    ser_rx_i : in  std_logic_vector(4 downto 0)

    );

end bridge_serdes_spartan6;

architecture rtl of bridge_serdes_spartan6 is

  component BUFPLL is
    port (
      IOCLK        : out std_ulogic;
      LOCK         : out std_ulogic;
      SERDESSTROBE : out std_ulogic;
      GCLK         : in  std_ulogic;
      LOCKED       : in  std_ulogic;
      PLLIN        : in  std_ulogic);
  end component BUFPLL;

  component ISERDES2 is
   
    port (
      CFB0      : out std_ulogic;
      CFB1      : out std_ulogic;
      DFB       : out std_ulogic;
      FABRICOUT : out std_ulogic;
      INCDEC    : out std_ulogic;
      Q1        : out std_ulogic;
      Q2        : out std_ulogic;
      Q3        : out std_ulogic;
      Q4        : out std_ulogic;
      SHIFTOUT  : out std_ulogic;
      VALID     : out std_ulogic;
      BITSLIP   : in  std_ulogic := 'L';
      CE0       : in  std_ulogic := 'H';
      CLK0      : in  std_ulogic;
      CLK1      : in  std_ulogic;
      CLKDIV    : in  std_ulogic;
      D         : in  std_ulogic;
      IOCE      : in  std_ulogic := 'H';
      RST       : in  std_ulogic := 'L';
      SHIFTIN   : in  std_ulogic);
  end component ISERDES2;

  component OSERDES2 is
    port (
      OQ        : out std_ulogic;
      SHIFTOUT1 : out std_ulogic;
      SHIFTOUT2 : out std_ulogic;
      SHIFTOUT3 : out std_ulogic;
      SHIFTOUT4 : out std_ulogic;
      TQ        : out std_ulogic;
      CLK0      : in  std_ulogic;
      CLK1      : in  std_ulogic;
      CLKDIV    : in  std_ulogic;
      D1        : in  std_ulogic;
      D2        : in  std_ulogic;
      D3        : in  std_ulogic;
      D4        : in  std_ulogic;
      IOCE      : in  std_ulogic := 'H';
      OCE       : in  std_ulogic := 'H';
      RST       : in  std_ulogic;
      SHIFTIN1  : in  std_ulogic;
      SHIFTIN2  : in  std_ulogic;
      SHIFTIN3  : in  std_ulogic;
      SHIFTIN4  : in  std_ulogic;
      T1        : in  std_ulogic;
      T2        : in  std_ulogic;
      T3        : in  std_ulogic;
      T4        : in  std_ulogic;
      TCE       : in  std_ulogic;
      TRAIN     : in  std_ulogic);
  end component OSERDES2;
  
  signal ser_clk_buf : std_logic;

  signal clk_125m_0 : std_logic;
  signal clk_500m_0 : std_logic;

  signal pllout_clk_fb     : std_logic;
  signal pllout_clk_125m_0 : std_logic;

  signal clk_500m_serdes : std_logic;
  signal serdes_strobe : std_logic;
  signal pll_locked : std_logic;

  type t_serdes_par_data is array(0 to 4) of std_logic_vector(3 downto 0);
  
  signal ser_rx_predelay, ser_rx_postdelay : std_logic_vector(4 downto 0);
  signal ser_tx_predelay, ser_tx_postdelay : std_logic_vector(4 downto 0);
  signal par_rx, par_tx : t_serdes_par_data;

  signal io_reset : std_logic;

  signal ser_clk_out_prebuf : std_logic;

  signal serdes_rx_bitslip : std_logic;

  signal BITSLIP_cnt : unsigned(3 downto 0);

 component chipscope_ila
    port (
      CONTROL : inout std_logic_vector(35 downto 0);
      CLK     : in    std_logic;
      TRIG0   : in    std_logic_vector(31 downto 0);
      TRIG1   : in    std_logic_vector(31 downto 0);
      TRIG2   : in    std_logic_vector(31 downto 0);
      TRIG3   : in    std_logic_vector(31 downto 0));
  end component;

  component chipscope_icon
    port (
      CONTROL0 : inout std_logic_vector (35 downto 0));
  end component;

  signal CONTROL                    : std_logic_vector(35 downto 0);
  signal CLK                        : std_logic;
  signal TRIG0                      : std_logic_vector(31 downto 0);
  signal TRIG1                      : std_logic_vector(31 downto 0);
  signal TRIG2                      : std_logic_vector(31 downto 0);
  signal TRIG3                      : std_logic_vector(31 downto 0);

  signal bridge_d_out : std_logic_vector(15 downto 0);
  signal bridge_frame_out : std_logic;

  signal cnt : unsigned(7 downto 0);
  signal cnt2 : unsigned(7 downto 0);
  
begin

  bridge_d_o <= bridge_d_out;
  bridge_frame_o <= bridge_frame_out;
  
    chipscope_ila_1 : chipscope_ila
    port map (
      CONTROL => CONTROL,
      CLK     => clk_125m_i,
      TRIG0   => TRIG0,
      TRIG1   => TRIG1,
      TRIG2   => TRIG2,
      TRIG3   => TRIG3);

    trig0(0) <= bridge_frame_i;
    trig0(16 downto 1) <= bridge_d_i;
    trig1(0) <= bridge_frame_out;
    trig1(16 downto 1) <= bridge_d_out;
    trig2(0) <= rst_n_i;
  trig3(7 downto 0) <= std_logic_vector(cnt);
  trig3(15 downto 8) <= std_logic_vector(cnt2);
  
  
  chipscope_icon_1 : chipscope_icon
    port map (
      CONTROL0 => CONTROL);


  process(clk_125m_i)
  begin
    if rising_edge(clk_125m_i) then
      cnt <= cnt + 1;
    end if;
  end process;

  process(clk_125m_0)
  begin
    if rising_edge(clk_125m_0) then
      cnt2 <= cnt2 + 1;
    end if;
  end process;
  
  
  io_reset <= not rst_n_i;


  gen1 : if g_mode = MASTER generate

    U_clk_buf : IBUFGDS
      generic map (
        DIFF_TERM => TRUE
        )
      port map (
        O  => ser_clk_buf,
        I  => ser_clk_p_i,
        IB => ser_clk_n_i
        );


    bridge_clk2_o <= ser_clk_buf;
    
    bridge_clk_o <= clk_125m_0;
    bridge_rst_n_o <= rst_n_i;
    
  end generate gen1;

  gen2 : if g_mode /= MASTER generate
    ser_clk_buf <= clk_125m_i;
  end generate gen2;

 gen_clk_test : if g_CLOCK_PERIOD = 1 generate

-- PLL for producing 125 MHz App clock
  U_App_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 8,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 4,          -- 250 MHz (IOSERDES)
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKOUT1_DIVIDE     => 16,          -- 62.5 MHz
      CLKOUT1_PHASE      => 0.000,
      CLKOUT1_DUTY_CYCLE => 0.500,
      CLKOUT2_DIVIDE     => 8,
      CLKOUT2_PHASE      => 180.000,    -- 125 MHz (180 deg)
      CLKOUT2_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 8.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => pllout_clk_fb,
      CLKOUT0  => clk_500m_0,
      CLKOUT1  => pllout_clk_125m_0,
      CLKOUT2  => open,                 --pllout_clk_sys,
      CLKOUT3  => open,
      CLKOUT4  => open,
      CLKOUT5  => open,
      LOCKED   => pll_locked,
      RST      => '0',
      CLKFBIN  => pllout_clk_fb,
      CLKIN    => ser_clk_buf);


  end generate gen_clk_test;


  gen_clk_125m : if g_CLOCK_PERIOD = 8 generate

-- PLL for producing 125 MHz App clock
  U_App_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 8,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 2,          -- 500 MHz (IOSERDES)
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKOUT1_DIVIDE     => 8,          -- 125 MHz
      CLKOUT1_PHASE      => 0.000,
      CLKOUT1_DUTY_CYCLE => 0.500,
      CLKOUT2_DIVIDE     => 8,
      CLKOUT2_PHASE      => 180.000,    -- 125 MHz (180 deg)
      CLKOUT2_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 8.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => pllout_clk_fb,
      CLKOUT0  => clk_500m_0,
      CLKOUT1  => pllout_clk_125m_0,
      CLKOUT2  => open,                 --pllout_clk_sys,
      CLKOUT3  => open,
      CLKOUT4  => open,
      CLKOUT5  => open,
      LOCKED   => pll_locked,
      RST      => '0',
      CLKFBIN  => pllout_clk_fb,
      CLKIN    => ser_clk_buf);


  end generate gen_clk_125m;

gen_clk_62m5 : if g_CLOCK_PERIOD = 16 generate

-- PLL for producing 125 MHz App clock
  U_App_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 16,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 4,          -- 250 MHz (IOSERDES)
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKOUT1_DIVIDE     => 16,          -- 62.5 MHz
      CLKOUT1_PHASE      => 0.000,
      CLKOUT1_DUTY_CYCLE => 0.500,
      CLKOUT2_DIVIDE     => 16,
      CLKOUT2_PHASE      => 180.000,    -- 62.5 MHz (180 deg)
      CLKOUT2_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 16.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => pllout_clk_fb,
      CLKOUT0  => clk_500m_0,
      CLKOUT1  => pllout_clk_125m_0,
      CLKOUT2  => open,                 --pllout_clk_sys,
      CLKOUT3  => open,
      CLKOUT4  => open,
      CLKOUT5  => open,
      LOCKED   => pll_locked,
      RST      => '0',
      CLKFBIN  => pllout_clk_fb,
      CLKIN    => ser_clk_buf);


  end generate gen_clk_62m5;
  
  
  
  U_clk_0_buf : BUFG
    port map (
      O => clk_125m_0,
      I => pllout_clk_125m_0);


  U_BufPLL : BUFPLL
    generic map (
      DIVIDE => 4
      )
    port map (
      PLLIN        => clk_500m_0,
      LOCKED       => pll_locked,
      GCLK         => clk_125m_0,
      SERDESSTROBE => serdes_strobe,
      IOCLK        => clk_500m_serdes,
      LOCK         => open
      );

  gen_serdeses : for i in 0 to 4 generate

    cmp_ibuf : IBUF
      generic map (
        IOSTANDARD => "LVCMOS33")
      port map (     
        I          => ser_rx_i(i),
        O          => ser_rx_predelay(i));

   cmp_obuf : OBUF
      generic map (
        IOSTANDARD => "LVCMOS33")
      port map (     
        O          => ser_tx_o(i),
        I          => ser_tx_postdelay(i));

    
    -- Instantiate the delay primitive
    -----------------------------------
  cmp_iodelay : IODELAY2
    generic map (
      DATA_RATE                => "SDR",
      IDELAY_VALUE             => g_SERDES_RX_DELAY_TAPS,
      IDELAY_TYPE              => "FIXED",
      COUNTER_WRAPAROUND       => "STAY_AT_LIMIT",
      DELAY_SRC                => "IDATAIN",
      SERDES_MODE              => "NONE",
      SIM_TAPDELAY_VALUE       => 75)
    port map (
      -- required datapath
      IDATAIN                => ser_rx_predelay(i),
      DATAOUT                => ser_rx_postdelay(i),
      T                      => '1',
      -- inactive data connections
      DATAOUT2               => open,
      DOUT                   => open,
      ODATAIN                => '0',
      TOUT                   =>  open,
       -- connect up the clocks
      IOCLK0                => '0',                 -- No calibration needed
      IOCLK1                => '0',                 -- No calibration needed
      -- Tie of the variable delay programming
      CLK                   => '0',
      CAL                   => '0',
      INC                   => '0',
      CE                    => '0',
      BUSY                  => open,
      RST                   => '0');



    
     cmp_iserdes : ISERDES2
       generic map (
         DATA_RATE      => "SDR",
         DATA_WIDTH     => 4,
         INTERFACE_TYPE => "RETIMED",
         SERDES_MODE    => "NONE",
         BITSLIP_ENABLE => true)
       port map (
         Q1         => par_rx(i)(3),
         Q2         => par_rx(i)(2),
         Q3         => par_rx(i)(1),
         Q4         => par_rx(i)(0),
         SHIFTOUT   => open,
         INCDEC     => open,
         VALID      => open,
         BITSLIP    => serdes_rx_bitslip,
         CE0        => '1',   
         CLK0       => clk_500m_serdes,
         CLK1       => '0',
         CLKDIV     => clk_125m_0,
         D          => ser_rx_postdelay(i),
         IOCE       => serdes_strobe,
         RST        => io_reset, 
         SHIFTIN    => '0' );

     cmp_oserdes : OSERDES2
       generic map (
         DATA_RATE_OQ   => "SDR",
         DATA_RATE_OT   => "SDR",
         TRAIN_PATTERN  => 0,
         DATA_WIDTH     => 4,
         SERDES_MODE    => "NONE",
         OUTPUT_MODE    => "SINGLE_ENDED")
       port map (
         D1         => par_tx(i)(3),
         D2         => par_tx(i)(2),
         D3         => par_tx(i)(1),
         D4         => par_tx(i)(0),
        T1         => '0',
        T2         => '0',
        T3         => '0',
        T4         => '0',
        SHIFTIN1   => '1',
        SHIFTIN2   => '1',
        SHIFTIN3   => '1',
        SHIFTIN4   => '1',
        SHIFTOUT1  => open,
        SHIFTOUT2  => open,
        SHIFTOUT3  => open,
        SHIFTOUT4  => open,
        TRAIN      => '0',
        OCE        => '1',
        CLK0       => clk_500m_serdes,
        CLK1       => '0',
        CLKDIV     => clk_125m_0,
        OQ         => ser_tx_postdelay(i),
        TQ         => open,
        IOCE       => serdes_strobe,
        TCE        => '1',
        RST        => IO_RESET);


    

    
  end generate gen_serdeses;


  gen3: if g_mode  = SLAVE generate
  -- cmp_oserdes_clk : OSERDES2
  --      generic map (
  --        DATA_RATE_OQ   => "SDR",
  --        DATA_RATE_OT   => "SDR",
  --        TRAIN_PATTERN  => 0,
  --        DATA_WIDTH     => 4,
  --        SERDES_MODE    => "NONE",
  --        OUTPUT_MODE    => "SINGLE_ENDED")
  --      port map (
  --        D1         => '0',
  --        D2         => '0',
  --        D3         => '1',
  --        D4         => '1',
  --       T1         => '0',
  --       T2         => '0',
  --       T3         => '0',
  --       T4         => '0',
  --       SHIFTIN1   => '1',
  --       SHIFTIN2   => '1',
  --       SHIFTIN3   => '1',
  --       SHIFTIN4   => '1',
  --       SHIFTOUT1  => open,
  --       SHIFTOUT2  => open,
  --       SHIFTOUT3  => open,
  --       SHIFTOUT4  => open,
  --       TRAIN      => '0',
  --       OCE        => '1',
  --       CLK0       => clk_500m_serdes,
  --       CLK1       => '0',
  --       CLKDIV     => clk_125m_0,
  --       OQ         => ser_clk_out_prebuf,
  --       TQ         => open,
  --       IOCE       => serdes_strobe,
  --       TCE        => '1',
  --       RST        => IO_RESET);


  -- OBUFDS_clk: OBUFDS
  --   generic map (
  --     IOSTANDARD  => "DEFAULT",
  --     SLEW        => "FAST")
  --   port map (
  --     O  => ser_clk_p_o,
  --     OB => ser_clk_n_o,
  --     I  => ser_clk_out_prebuf);
  
  end generate gen3;

  gen_serdes_data : for i in 0 to 3 generate
    par_tx(i)(0) <= bridge_d_i(i*4+0);
    par_tx(i)(1) <= bridge_d_i(i*4+1);
    par_tx(i)(2) <= bridge_d_i(i*4+2);
    par_tx(i)(3) <= bridge_d_i(i*4+3);
    bridge_d_out(i*4+0) <= par_rx(i)(0);
    bridge_d_out(i*4+1) <= par_rx(i)(1);
    bridge_d_out(i*4+2) <= par_rx(i)(2);
    bridge_d_out(i*4+3) <= par_rx(i)(3);

    bridge_frame_out <= par_rx(4)(0);
    par_tx(4)(0) <= bridge_frame_i;
    par_tx(4)(1) <= '0';
    par_tx(4)(2) <= '0';
    par_tx(4)(3) <= '0';
    
    
    
  end generate gen_serdes_data;
  

  p_serdes_bitslip : process(clk_125m_0, rst_n_i)
  begin
    if rst_n_i = '0' then
      BITSLIP_cnt <= (others => '0');
      serdes_rx_bitslip <= '0';
    elsif rising_edge(clk_125m_0) then
      if BITSLIP_cnt = 0 then
        if par_rx(4) /= "0000" and par_rx(4) /= "0001" then
          serdes_rx_bitslip <= '1';
          BITSLIP_cnt <= to_unsigned(10, 4);
        end if;
      else
        serdes_rx_bitslip <= '0';
        BITSLIP_cnt <= BITSLIP_cnt - 1;
      end if;
    end if;
  end process;
  
  
end rtl;

